from django.views.generic import View
from django.views.decorators.csrf import csrf_exempt


class Resource(View):

    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        # If the HTTP method is PUT, load the POST data
        if request.method.lower() == 'put':
            request.method = 'POST'
            request._load_post_and_files()
            request.method = 'PUT'
        else:
            request._load_post_and_files()

        # Check for an overriding method in the GET or POST data,
        # and set the HTTP method on the request appropriately
        if request.GET and request.GET.get('_method', None):
            GET = request.GET.copy()
            request.method = GET.pop('_method')[0].upper()
            request.GET = GET
        elif request.POST and request.POST.get('_method', None):
            POST = request.POST.copy()
            request.method = POST.pop('_method')[0].upper()
            request.POST = POST

        # Manually switch POST data to PUT
        if (request.method == 'PUT'):
            request.PUT = request.POST

        # Try to dispatch to the right method; if a method doesn't exist,
        # defer to the error handler. Also defer to the error handler if the
        # request method isn't on the approved list.
        if request.method.lower() in self.http_method_names:
            handler = getattr(self, request.method.lower(), self.http_method_not_allowed)
        else:
            handler = self.http_method_not_allowed

        # Have to set request on the View object to make http_method_not_allowed happy
        self.request = request

        return handler(request, *args, **kwargs)



import time
import hmac
import hashlib


def calculate_signature(key, data, timestamp=None):
    """
    Calculates the signature for the given request data.
    """
    # Create a timestamp if one was not given
    if timestamp is None:
        timestamp = int(time.time())

    # Construct the message from the timestamp and the data in the request
    message = str(timestamp) + ''.join("%s%s" % (k,v) for k,v in sorted(data.items()))

    # Calculate the signature (HMAC SHA256) according to RFC 2104
    signature = hmac.HMAC(str(key), message, hashlib.sha256).hexdigest()

    return signature



import collections
import mimetypes

import mimeparse

from django.conf import settings
from django.shortcuts import render_to_response
from django.http import HttpResponse
from django.core import serializers
from django.db.models.query import QuerySet

from .utils.decorators import wrap_object

try:
    import simplejson as json
except ImportError:
    import json


def serialize_json(content):
    if isinstance(content, QuerySet):
        json_serializer = serializers.get_serializer('json')()
        serialized_content = json_serializer.serialize(content, ensure_ascii=False)
    else:
        json_encoder = serializers.json.DjangoJSONEncoder
        serialized_content = json.dumps(content, cls=json_encoder, ensure_ascii=False)
    return serialized_content

DEFAULT_MIMETYPE = 'application/json'
SUPPORTED_MIMETYPES = {
    'application/json': serialize_json
}


class RESTfulResponse(object):
    """
    Can be used as a decorator or an instance to properly formatted content

    This class provides content negotiation for RESTful requests. The content-
    type of a response is determined by the ACCEPT header of the reqeust or by
    an overriding file extension in the URL (e.g., path/to/some/resource.xml)
    will return the response formatted as XML.

    This class creates an instance that can be used directly to transform a
    python dict into a properly formatted response via the render_to_response
    method or it can be used as a decorator for class-based views or for an
    individual method within a class-based view. If a requested mimetype is
    not found amongst the supported mimetypes, the content-type of the response
    will default to 'application/json'.

    This class is inspired by an excellent blog post from James Bennett. See
    http://www.b-list.org/weblog/2008/nov/29/multiresponse/ for more
    information.
    """
    def __init__(self, mimetype_mapping=None):
        self.supported_mimetypes = SUPPORTED_MIMETYPES.copy()
        if mimetype_mapping:
            self.supported_mimetypes.update(mimetype_mapping)

    def __call__(self, view_obj):
        def decorator(view_func):
            def wrapper(request, *args, **kwargs):
                results = view_func(request, *args, **kwargs)

                # Get the status code, if one was provided
                if isinstance(results, collections.Sequence) and len(results) == 2:
                    try:
                        data, status_code = results[0], int(results[1])
                    except Exception:
                        data, status_code = results, 200
                else:
                    data, status_code = results, 200

                # TODO: What about a view that returns a normal Django
                #       HttpResponse object? Should we allow this to error
                #       out when one is encountered? Should we just allow it
                #       to propogate through (i.e., kind of an override)? Or,
                #       should we fectch the content from it and pass that to
                #       render_to_response method?
                response = self.render_to_response(request, data, status_code)
                return response
            return wrapper
        return wrap_object(view_obj, decorator)

    def render_to_response(self, request, data=None, status=200):
        mimetype = mimeparse.best_match(self.supported_mimetypes.keys(), request.META['HTTP_ACCEPT'])
        mimetype = mimetypes.guess_type(request.path_info.rstrip('/'))[0] or mimetype
        content_type = '%s; charset=%s' % (mimetype, settings.DEFAULT_CHARSET)

        templ_or_func = self.supported_mimetypes.get(mimetype)

        # If a template or function isn't found, return a 415 (unsupportted media type) response
        if not templ_or_func:
            return HttpResponse(status=415)

        if isinstance(templ_or_func, str):
            def serialize(data):
                data = data or {}
                response = render_to_response(templ_or_func, data)
                response['Content-Type'] = content_type
                response.status_code = status
                return response
        else:
            def serialize(data):
                if data:
                    response = HttpResponse(templ_or_func(data), content_type=content_type, status=status)
                else:
                    response = HttpResponse(content_type=content_type, status=status)
                return response

        return serialize(data)



from optparse import make_option
import datetime
import time
import urllib

from django.core.management.base import BaseCommand
from ...auth.signature import calculate_signature


class Command(BaseCommand):
    help = """URL encode the given data.

Returns a URL encoded string for the given set of data. If a secret key is
specified, the secure signature is also calculated and added to the result."""

    args = "secret_key [key=value key=value ...]"

    option_list = BaseCommand.option_list + (
        make_option('--secret-key',
            dest='secret-key',
            action='store',
            help='Calculate the secure signature with the secret key'),
        )

    def handle(self, *data, **options):
        # Convert the data from a list of key, value pairs to a dict
        data = dict(item.split('=') for item in data)

        # Get the secret key if one was provided
        secret_key = options.get('secret-key', None)

        if secret_key:
            # If the timestamp was specified, use it and remove it from the data
            # before calculating the signature, otherwise, the time use will be
            # now in UTC time.
            timestamp = data.pop('t', None)
            if timestamp:
                timestamp = int(timestamp)
            else:
                dt = datetime.datetime.utcnow()
                timestamp = int(time.mktime(dt.timetuple()))
            signature = calculate_signature(secret_key, data, timestamp)
            data['t'] = timestamp
            data['sig'] = signature

        print urllib.urlencode(data)



from django.forms import ModelForm as DjangoModelForm
from django.core.exceptions import ObjectDoesNotExist
from django.db.models.fields import NOT_PROVIDED


class ModelForm(DjangoModelForm):
    """
    Allows a model form to validate properly with default attributes missing.

    Django model forms expect all attributes for a given model to be present
    in a request. Since a typical request comes from an HTML form and the HTML
    for that form is typically generated by the form object (which sets all
    of the HTML form values to the defaults of the model), this is usually not
    an issue. However, with RESTful requests, we cannot assume that all
    attributes will be present in the request.

    This class allows the form to validate even if some (optional) attributes
    are not present in the original request. It does so by adding the default
    values (from the model) to the data from the request for each missing
    attribute. It must add these elements to the data object, as opposed to
    just passing it to the form's __init__ method as the 'initial' parameter
    since 'initial' is only used to render the HTML form and not for the
    validation of the data.

    UPDATE:
    This mixin has been updated to also populate form with values from an
    instance, if one has been provided.
    Note: this has not yet been tested, and likely does not work, on
    many-to-many relationships.
    """

    def __init__(self, *args, **kwargs):
        super(ModelForm, self).__init__(*args, **kwargs)
        for field in self.instance.__class__._meta.fields:
            if field.name not in self.data.keys() or self.data.get(field.name) == None:
                inst_val = None
                if self.instance.id:
                    try:
                        inst_val = getattr(self.instance, field.name)
                        #Get the ID if this is a model object and
                        #we're dealing with a foreign key
                        #relationship.
                        inst_val = getattr(inst_val, 'id')
                    except (AttributeError, ObjectDoesNotExist):
                        pass
                    self.data[field.name] = inst_val
                elif field.default != NOT_PROVIDED:
                    self.data[field.name] = field.default




from datetime import datetime

from .signature import calculate_signature
from ..utils.decorators import wrap_object
from ..exceptions import HttpError


def auth_required(secret_key_func):
    """
    Requires that the user be authenticated either by a signature or by
    being actively logged in.
    """
    def actual_decorator(obj):

        def test_func(request, *args, **kwargs):
            secret_key = secret_key_func(request, *args, **kwargs)
            return validate_signature(request, secret_key) or request.user.is_authenticated()

        decorator = request_passes_test(test_func)
        return wrap_object(obj, decorator)

    return actual_decorator


def login_required(obj):  # TODO: Add the redirect_field_name and login_url fields
    """
    Requires that the user be logged in order to gain access to the resource
    at the specified the URI.
    """
    decorator = request_passes_test(lambda r, *args, **kwargs: r.user.is_authenticated())
    return wrap_object(obj, decorator)


def admin_required(obj):
    """
    Requires that the user be logged AND be set as a superuser
    """
    decorator = request_passes_test(lambda r, *args, **kwargs: r.user.is_superuser)
    return wrap_object(obj, decorator)


def signature_required(secret_key_func):
    """
    Requires that the request contain a valid signature to gain access
    to a specified resource.
    """
    def actual_decorator(obj):

        def test_func(request, *args, **kwargs):
            secret_key = secret_key_func(request, *args, **kwargs)
            return validate_signature(request, secret_key)

        decorator = request_passes_test(test_func)
        return wrap_object(obj, decorator)

    return actual_decorator


def request_passes_test(test_func):
    """
    Decorator for resources that checks that the request passes the given test.
    If the request fails the test a 401 (Unauthorized) response is returned,
    otherwise the view is executed normally. The test should be a callable that
    takes an HttpRequest object and any number of positional and keyword
    arguments as defined by the urlconf entry for the decorated resource.
    """
    def decorator(view_func):
        # @wraps(view_func, assigned=available_attrs(view_func))
        def _wrapped_view(request, *args, **kwargs):
            if test_func(request, *args, **kwargs):
                return view_func(request, *args, **kwargs)
            raise HttpError(status=401)
        return _wrapped_view
    return decorator


def validate_signature(request, secret_key):
    """
    Validates the signature associated with the given request.
    """

    # Extract the request parameters according to the HTTP method
    data = request.GET.copy()
    if request.method.lower() == 'post':
        data.update(request.POST.copy())
    elif request.method.lower() == 'put':
        data.update(request.PUT.copy())

    # Make sure the request contains a signature
    if data.get('sig', False):
        sig = data['sig']
        del data['sig']
    else:
        return False

    # Make sure the request contains a timestamp
    if data.get('t', False):
        timestamp = int(data.get('t', False))
        del data['t']
    else:
        return False

    # Make sure the signature has not expired
    delta = datetime.utcnow() - datetime.utcfromtimestamp(timestamp)
    if delta.seconds > 5 * 60:  # If the signature is older than 5 minutes, it's invalid
        return False

    # Make sure the signature is valid
    return sig == calculate_signature(secret_key, data, timestamp)



import inspect
from functools import update_wrapper

from django.utils.decorators import method_decorator, available_attrs


def wrap_object(obj, decorator):
    """
    Decorates the given object with the decorator function.

    If obj is a method, the method is decorated with the decorator function
    and returned. If obj is a class (i.e., a class based view), the methods
    in the class corresponding to HTTP methods will be decorated and the
    resultant class object will be returned.
    """
    actual_decorator = method_decorator(decorator)

    if inspect.isfunction(obj):
        wrapped_obj = actual_decorator(obj)
        update_wrapper(wrapped_obj, obj, assigned=available_attrs(obj))
    elif inspect.isclass(obj):
        for method_name in obj.http_method_names:
            if hasattr(obj, method_name):
                method = getattr(obj, method_name)
                wrapped_method = actual_decorator(method)
                update_wrapper(wrapped_method, method, assigned=available_attrs(method))
                setattr(obj, method_name, method)
        wrapped_obj = obj
    else:
        raise TypeError("received an object of type '{0}' expected 'function' or 'classobj'.".format(type(obj)))

    return wrapped_obj



from django.http import HttpResponse

from .response import RESTfulResponse


class HttpError(Exception):
    def __init__(self, message=None, status=500):
        super(HttpError, self).__init__(message)
        self.status = status

    def __repr__(self):
        return 'HttpError(%r, %r)' % (self.status, self.message)


class ExceptionMiddleware(object):
    def process_exception(self, request, exception):
        """
        Returns the proper HttpRespone if an HttpError was thrown
        """
        response = None
        if isinstance(exception, HttpError):
            if exception.message:
                context_dict = {'message': exception.message}
                response = RESTfulResponse().render_to_response(request, context_dict, exception.status)
            else:
                response = RESTfulResponse().render_to_response(request, status=exception.status)
        return response



try:
    from setuptools import setup, find_packages
except ImportError:
    import sys
    print >>sys.stderr, "Error: setuptools is required to install this package."
    sys.exit(1)


__version__ = '0.4.5'
__author__ = 'Christopher Roach'
__email__ = 'croach@freshplum.com'
__license__ = 'MIT'

setup(
    name='django-simple-rest',
    version=__version__,
    author=__author__,
    author_email=__email__,
    description='A drop dead simple package for creating RESTful APIs on top of Django',
    long_description=open('README.rst').read(),
    url='https://github.com/freshplum/django-simple-rest',
    packages=find_packages(),
    install_requires=['setuptools', 'mimeparse'],
    zip_safe=False,
    keywords='rest,django',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)
